/*****************************************************************************
 * Filename:    cryptfs/actions.c                                            *
 * Description: Handles the actions and does fsck/mounting                   *
 * Copyright:   2005 by Alexander Motzkau                                    *
 *****************************************************************************/

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>

#include "actions.h"
#include "key.h"
#include "misc.h"
#include "dm.h"
#include "vt.h"

enum bool handle_command(struct command *cmd, enum bool success, struct configuration* config)
{
    pid_t child;

    if((success && cmd->on_state == on_failure) ||
       (!success && cmd->on_state == on_success))
        return true;
    
    if(cmd->needtty)
	get_tty();

    child = fork();
    
    if(!child)
    {
	if(cmd->uid != NULL)
	    setuidnam(cmd->uid);
	
	execl("/bin/sh", "/bin/sh", "-c", cmd->cmd, NULL);
	errnomsg("execl");
	
	exit(1);
    }
    else if(child<0)
    {
	errnomsg("fork");
	return false;
    }
    else
    {
	int status;
	
	if(waitpid(child, &status, 0)<0)
	{
	    errnomsg("waitpid");
	    return false;
	}
	
	if(WIFEXITED(status) && WEXITSTATUS(status)==0)
	    return true;
	
	if(cmd->ignore_status)
	    return true;
	
	return false;
    }

    return false; /* Not reached. */
}

enum bool handle_commands(struct command *cmd, enum bool success, struct configuration* config)
{
    while(cmd != NULL)
    {
	success = handle_command(cmd, success, config) && success;
	cmd = cmd->next;
    }
    
    return success;
}

enum bool handle_action(char* action, struct configuration* config)
{
    /* Let's go for it. */
    
    struct action* ac;
    struct map* mapping;
    struct key* key;
    enum bool success;
    
    ac = find_action(action, config);

    /* Check, whether there's anything todo. */    
    success = false;
    for(mapping = ac->mappings; mapping != NULL; mapping = mapping->next)
    {
	unsigned long bsize;
	char *dmname;
	
	dmname = get_crypt_name(mapping);
	bsize = 0;
	
	if(get_blockdev_size(dmname, &bsize)?(bsize==0):(errno!=EACCES))
	{
	    mapping->done = false;
	    success = true;
	}
	else
	    mapping->done = true;
	
        xfree(dmname);
    }
    
    if(!success)
	return true; /* Nothing to do. */
    
    if(!handle_commands(ac->preexec, true, config))
    {
	handle_commands(ac->postexec, false, config);
	return false;
    }
    
    key = ac->key;
    success = false;

    for(mapping = ac->mappings; mapping != NULL; mapping = mapping->next)
    {
	char* dmname;
	void *k;
	int klen;
	
	k = NULL;
	
	if(mapping->done)
	    continue;
	
	if(mapping->key == NULL && (k=handle_key(key, &klen, config))==NULL)
	{
	    handle_commands(ac->postexec, false, config);
	    return false;
	}
	
	if(!handle_commands(mapping->preexec, true, config))
	{
	    handle_commands(mapping->postexec, false, config);
	    continue;
	}
	
	if(mapping->key != NULL && (k=handle_key(mapping->key, &klen, config))==NULL)
	{
	    handle_commands(mapping->postexec, false, config);
	    continue;
	}
	
	if(!setup_crypt(mapping, k, klen, config))
	{
	    handle_commands(mapping->postexec, false, config);
	    continue;
	}
	
	dmname = get_crypt_name(mapping);
	
	if(mapping->fsck)
	{
	    pid_t child;
	    
	    get_tty();
	    
	    child = fork();

	    if(!child)
	    {
		/* fsck -n [ -t mapping->fs ] devname */
		char* fsck_arg[5];
		int a;
	    
		fsck_arg[0] = "/sbin/fsck";
		a = 1;
	    
		if(mapping->fs != NULL && strlen(mapping->fs))
		{
		    fsck_arg[a++] = "-t";
		    fsck_arg[a++] = mapping->fs;
		}
		fsck_arg[a++] = dmname;
		fsck_arg[a++] = NULL;
		
		execv("/sbin/fsck", fsck_arg);
		errnomsg("execv /sbin/fsck");
		exit(1);
	    }
	    else if(child<0)
	    {
		errnomsg("fork");
		remove_crypt(mapping, config);
		handle_commands(mapping->postexec, false, config);
		continue;
	    }
	    else
	    {
		int status;
		
		if(waitpid(child, &status, 0)<0)
		{
		    errnomsg("waitpid");
		    remove_crypt(mapping, config);
		    handle_commands(mapping->postexec, false, config);
		    continue;
		}
	
		if(!WIFEXITED(status) || WEXITSTATUS(status)!=0)
		{
		    errmsg("%s: fsck failed.\n", dmname);
		    remove_crypt(mapping, config);
		    handle_commands(mapping->postexec, false, config);
		    continue;
		}
	    }
	}
	
	if(mapping->mount)
	{
	    pid_t child;
	    
	    child = fork();
	    
	    if(!child)
	    {
		/* mount [ -t mapping->fs ] [ -o mapping->options ]
			 devname [ mapping->point] */
		
		char* mnt_arg[7];
		int a;
	    
		mnt_arg[0] = "/bin/mount";
		a = 1;
	    
		if(mapping->fs != NULL && strlen(mapping->fs))
		{
		    mnt_arg[a++] = "-t";
		    mnt_arg[a++] = mapping->fs;
		}

		if(mapping->options != NULL && strlen(mapping->options))
		{
		    mnt_arg[a++] = "-o";
		    mnt_arg[a++] = mapping->options;
		}
		mnt_arg[a++] = dmname;
		if(mapping->point != NULL && strlen(mapping->point))
		    mnt_arg[a++] = mapping->point;
		mnt_arg[a++] = NULL;
		
		execv("/bin/mount", mnt_arg);
		errnomsg("execv /bin/mount");
		exit(1);
	    }
	    else if(child<0)
	    {
		errnomsg("fork");
		remove_crypt(mapping, config);
		handle_commands(mapping->postexec, false, config);
		continue;
	    }
	    else
	    {
		int status;
		
		if(waitpid(child, &status, 0)<0)
		{
		    errnomsg("waitpid");
		    remove_crypt(mapping, config);
		    handle_commands(mapping->postexec, false, config);
		    continue;
		}

		if(!WIFEXITED(status) || WEXITSTATUS(status)!=0)
		{
		    errmsg("%s: mount failed.\n", dmname);
		    remove_crypt(mapping, config);
		    handle_commands(mapping->postexec, false, config);
		    continue;
		}
	    }
	}
	
	if(!handle_commands(mapping->postexec, true, config))
	{
	    remove_crypt(mapping, config);
	    continue;
	}

	success = true; /* At least one successful mapping. */
    }

    handle_commands(ac->postexec, success, config);
    
    return true;
}
