/*****************************************************************************
 * Filename:                                                                 *
 * Description:                                                              *
 * Copyright:   2005 by Alexander Motzkau                                    *
 *****************************************************************************/
#include <getopt.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>

#include "actions.h"
#include "misc.h"
#include "xml.h"

#define DEFAULT_XML "/etc/dmcrypt.xml"

struct actionname
{
    char* name;
    struct actionname* next;
};

struct main_options
{
    char* xmlfile;
    struct actionname* actions;
};

void free_main_options(struct main_options* opts)
{
    struct actionname* ac;
    
    xfree(opts->xmlfile);

    ac = opts->actions;
    while(ac!=NULL)
    {
	struct actionname* temp;
	temp = ac;
	ac = ac->next;
	xfree(temp->name);
	xfree(temp);
    };
    
    xfree(opts);
};

struct main_options* parse_options(int argc, char* argv[])
{
    struct main_options* result;
    
    result = (struct main_options*)xmalloc(sizeof(struct main_options));
    result->xmlfile = NULL;
    result->actions = NULL;
    
    while(1)
    {
	int o;
	static struct option long_opts[] =
	    {
		{ "filename", 1, NULL, 'f'},
		{ "help", 0, NULL, 'h'},
		{ NULL, 0, NULL, 0 },
	    };
	    
	o = getopt_long(argc, argv, "f:h", long_opts, NULL);
	if(o == -1)
	    break;
	
	switch(o)
	{
	    case '?':
		free_main_options(result);
		return NULL;
	    case 'f':
		if(result->xmlfile != NULL)
		{
		    fprintf(stderr, "filename given twice.\n");
		    free_main_options(result);
		    return NULL;
		}
		result->xmlfile = xmallocstr(optarg);
		break;
	    case 'h':
		printf("Usage: dmcrypt [options] [action ...]\n\n"
		       "Options:\n"
		       "  -f, --filename=XMLFILE  File containing the definitions.\n"
		       "                          Default: /etc/dmcrypt.xml\n"
		       "  -h, --help              Show options.\n");
		free_main_options(result);
		return NULL;
	}
    }
    
    if(result->xmlfile==NULL)
	result->xmlfile = xmallocstr(DEFAULT_XML);

    while(optind < argc)
    {
	struct actionname* action;
	action = (struct actionname*) xmalloc(sizeof(struct actionname));
	
	action->name = xmallocstr(argv[--argc]);
	action->next = result->actions;
	result->actions = action;
    }
    
    return result;
};

int main(int argc, char* argv[])
{
    struct configuration *config;
    struct main_options* opts;
    struct actionname* ac;

    setlocale(LC_CTYPE, "");
        
    opts = parse_options(argc, argv);
    if(opts == NULL)
	return 1;
        
    config = parse_file(opts->xmlfile);
    if(config == NULL)
	return 1;

    if(!complete_config(config))
	return 1;

    if(opts->actions == NULL) /* No action given. */
    {
	struct action *act;
	for(act = config->actions; act != NULL; act = act->next)
	    printf("%s\n", act->name);
	return 0;
    }

    for(ac = opts->actions; ac != NULL; ac = ac->next)
	if(find_action(ac->name, config)==NULL)
	{
	    fprintf(stderr, "Action '%s' undefined.\n", ac->name);
	    return 1;
	}
    
    for(ac = opts->actions; ac != NULL; ac = ac->next)
	if(!handle_action(ac->name, config))
	    fprintf(stderr, "Action '%s' failed.\n", ac->name);
    
    return 0;
}
